﻿using System;
using System.ComponentModel;
using System.Globalization;

namespace Petzold.Controls
{
    public class DatePresenter : INotifyPropertyChanged
    {
        const int MINYEAR = 1601;
        const int MAXYEAR = 3000;

        public event PropertyChangedEventHandler PropertyChanged;

        YearInfo selectedYear;
        MonthInfo selectedMonth;
        DayInfo selectedDay;

        public DatePresenter()
        {
            AllYears = new YearInfo[MAXYEAR - MINYEAR + 1];

            for (int year = 0; year < AllYears.Length; year++)
            {
                AllYears[year] = new YearInfo
                {
                    YearNumber = year + MINYEAR
                };
            }

            AllMonths = new MonthInfo[12];

            for (int month = 0; month < 12; month++)
                AllMonths[month] = new MonthInfo
                    {
                        MonthNumber = month + 1
                    };

            DateTime = DateTime.Now;
        }

        public YearInfo[] AllYears { protected set; get; }

        public MonthInfo[] AllMonths { protected set; get; }

        public DayInfo[] AllDays { protected set; get; }

        public DateTime DateTime
        {
            set
            {
                if (SelectedYear == null || SelectedMonth == null || SelectedDay == null ||
                    value.Year != SelectedYear.YearNumber || value.Month != SelectedMonth.MonthNumber || value.Day != SelectedDay.DayNumber)
                {
                    SelectedYear = AllYears[value.Year - MINYEAR];
                    SelectedMonth = AllMonths[value.Month - 1];
                    SelectedDay = AllDays[value.Day - 1];

                    OnPropertyChanged(new PropertyChangedEventArgs("DateTime"));
                }
            }
            get
            {
                if (SelectedYear != null && SelectedMonth != null && SelectedDay != null)
                    return new DateTime(SelectedYear.YearNumber, SelectedMonth.MonthNumber, SelectedDay.DayNumber);

                return new DateTime();
            }
        }

        public DayInfo SelectedDay
        {
            set
            {
                if (value != selectedDay)
                {
                    selectedDay = value;
                    OnPropertyChanged(new PropertyChangedEventArgs("SelectedDay"));
                    OnPropertyChanged(new PropertyChangedEventArgs("DateTime"));
                }
            }
            get
            {
                return selectedDay;
            }
        }

        public MonthInfo SelectedMonth 
        {
            set
            {
                if (value != selectedMonth)
                {
                    selectedMonth = value;

                    if (SelectedYear != null && value != null)
                        MakeNewDays(SelectedYear.YearNumber, value.MonthNumber);

                    OnPropertyChanged(new PropertyChangedEventArgs("SelectedMonth"));
                    OnPropertyChanged(new PropertyChangedEventArgs("DateTime"));
                }
            }

            get
            {
                return selectedMonth;
            }
        }

        public YearInfo SelectedYear
        {
            set
            {
                if (value != selectedYear)
                {
                    selectedYear = value;

                    if (value != null && SelectedMonth != null)
                        MakeNewDays(value.YearNumber, SelectedMonth.MonthNumber);

                    OnPropertyChanged(new PropertyChangedEventArgs("SelectedYear"));
                    OnPropertyChanged(new PropertyChangedEventArgs("DateTime"));
                }
            }
            get
            {
                return selectedYear;
            }
        }

        void MakeNewDays(int year, int month)
        {
            int daysInMonth = DateTime.DaysInMonth(year, month);
            AllDays = new DayInfo[daysInMonth];

            for (int day = 0; day < daysInMonth; day++)
                AllDays[day] = new DayInfo
                {
                    Year = year,
                    Month = month,
                    DayNumber = day + 1
                };

            OnPropertyChanged(new PropertyChangedEventArgs("AllDays"));

            if (SelectedDay != null)
                SelectedDay = AllDays[Math.Min(SelectedDay.DayNumber, AllDays.Length) - 1];
        }

        void OnPropertyChanged(PropertyChangedEventArgs args)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, args);
        }
    }
}
