﻿using System;
using System.ComponentModel;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;

namespace Petzold.Controls
{
    public partial class WindowedItemsControl: ItemsControl
    {
        bool isWrappableStackPanel;
        int numberItems;
        double ItemsHeight;
        double ViewportHeight;
        double ExtentHeight;

        public event DependencyPropertyChangedEventHandler IsActiveChanged;

        public static readonly DependencyProperty IsActiveProperty =
            DependencyProperty.Register("IsActive",
                typeof(bool),
                typeof(WindowedItemsControl),
                new PropertyMetadata(false, OnIsActiveChanged));

        public static readonly DependencyProperty SelectedIndexProperty = 
            Selector.SelectedIndexProperty.AddOwner(
                typeof(WindowedItemsControl), 
                new FrameworkPropertyMetadata(-1, OnSelectedIndexChanged));

        public static readonly DependencyProperty SelectedItemProperty =
            Selector.SelectedItemProperty.AddOwner(
                typeof(WindowedItemsControl),
                new FrameworkPropertyMetadata(null, OnSelectedItemChanged));

        static readonly DependencyPropertyKey VerticalOffsetPropertyKey =
            DependencyProperty.RegisterReadOnly("VerticalOffset",
                typeof(double),
                typeof(WindowedItemsControl),
                new PropertyMetadata(OnVerticalOffsetChanged));

        public static readonly DependencyProperty VerticalOffsetProperty =
            VerticalOffsetPropertyKey.DependencyProperty;

        public static readonly DependencyProperty StartIndexProperty =
            DependencyProperty.RegisterAttached("StartIndex",
                typeof(double),
                typeof(WindowedItemsControl),
                new FrameworkPropertyMetadata(0.0, 
                    FrameworkPropertyMetadataOptions.Inherits));

        public WindowedItemsControl()
        {
            IsManipulationEnabled = true;
        }

        public bool IsActive
        {
            set { SetValue(IsActiveProperty, value); }
            get { return (bool)GetValue(IsActiveProperty); }
        }

        public int SelectedIndex
        {
            set { SetValue(SelectedIndexProperty, value); }
            get { return (int)GetValue(SelectedIndexProperty); }
        }

        public object SelectedItem
        {
            set { SetValue(SelectedItemProperty, value); }
            get { return (object)GetValue(SelectedItemProperty); }
        }

        public double VerticalOffset
        {
            protected set { SetValue(VerticalOffsetPropertyKey, value); }
            get { return (double)GetValue(VerticalOffsetProperty); }
        }

        public double StartIndex
        {
            set { SetValue(StartIndexProperty, value); }
            get { return (double)GetValue(StartIndexProperty); }
        }

        // IsActive property-changed handlers
        static void OnIsActiveChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as WindowedItemsControl).OnIsActiveChanged(args);
        }

        void OnIsActiveChanged(DependencyPropertyChangedEventArgs args)
        {
            if (IsActiveChanged != null)
                IsActiveChanged(this, args);
        }

        // SelectedIndex property-changed handlers
        static void OnSelectedIndexChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as WindowedItemsControl).OnSelectedIndexChanged(args);
        }

        void OnSelectedIndexChanged(DependencyPropertyChangedEventArgs args)
        {
            if (SelectedIndex >= 0 && SelectedIndex < this.Items.Count)
                SelectedItem = this.Items[SelectedIndex];
            else
                SelectedItem = null;

            SetVerticalOffsetFromSelectedIndex();
        }

        // SelectedItem property-changed handlers
        static void OnSelectedItemChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as WindowedItemsControl).OnSelectedItemChanged(args);
        }

        void OnSelectedItemChanged(DependencyPropertyChangedEventArgs args)
        {
            if (SelectedItem == null)
                SelectedIndex = -1;
            else
                SelectedIndex = this.Items.IndexOf(SelectedItem);
        }

        static void OnVerticalOffsetChanged(DependencyObject obj, DependencyPropertyChangedEventArgs args)
        {
            (obj as WindowedItemsControl).OnVerticalOffsetChanged(args);
        }

        void OnVerticalOffsetChanged(DependencyPropertyChangedEventArgs args)
        {
            if (!isWrappableStackPanel)
            {
                InvalidateArrange();
            }
            else if (ItemsHeight > 0 && numberItems > 0)
            {
                double startIndex = VerticalOffset / ItemsHeight;

                startIndex %= numberItems;
                startIndex = (startIndex + numberItems) % numberItems;
                StartIndex = startIndex;
            }
        }

        void SetVerticalOffsetFromSelectedIndex()
        {
            numberItems = this.Items.Count;
            ItemsHeight = numberItems > 0 ? ExtentHeight / numberItems : 0;

            if (SelectedIndex != -1 && !manipulationInProgress)
            {
                VerticalOffset = VerticalOffsetFromCenteredIndex(SelectedIndex);
            }
        }

        double VerticalOffsetFromCenteredIndex(int index)
        {
            return index * ItemsHeight - (ViewportHeight - ItemsHeight) / 2;
        }

        int NormalizeIndex(int index)
        {
            if (numberItems == 0)
                return 0;

            while (index < 0)
                index += numberItems;

            while (index >= numberItems)
                index -= numberItems;

            return index;
        }

        // not used yet
        double NormalizeIndex(double index)
        {
            if (numberItems == 0)
                return 0;

            while (index < 0)
                index += numberItems;

            while (index >= numberItems)
                index -= numberItems;

            return index;
        }
        
        double FractionalCenteredIndexFromVerticalOffset(double verticalOffset, bool normalize)
        {
            double index = (verticalOffset + (ViewportHeight - ItemsHeight) / 2) / ItemsHeight;

            if (normalize)
            {
                index %= numberItems;
                index = (index + numberItems) % numberItems;
            }

            return index;
        }
    }
}
